#include "tcpip.h"

unsigned int num_tcp_head=0;
unsigned int num_tcp_rules=0;

/* tcp_rules_arr[state][dir] */
struct tcp_head *tcp_rules_arr[3][3]={
	{NULL, NULL, NULL},
	{NULL, NULL, NULL},
	{NULL, NULL, NULL}
};

char *tcp_ignore_words[]={
	"stateless",
	"flow",
	NULL
};

/* Generator */
struct generator tcp_gen=init_generator("sig.tcp");

int tcp_do_match(struct tcp_head *h, struct packet *p, unsigned int l)
{
	struct pkt_iphdr *iph;
	struct pkt_tcphdr *tcph;
	u_int16_t sp, dp;

	iph=p->layer[l].h.ip;
	tcph=p->layer[l+1].h.tcp;
	sp=ntohs(tcph->sport);
	dp=ntohs(tcph->dport);

	for(; h; h=h->next) {
		if ( 
			(((h->saddr&h->smask) == (iph->saddr&h->smask))
			^ ((h->flags&NEGATE_SADDR)!=0) ) &&
			
			(((h->daddr&h->dmask) == (iph->daddr&h->dmask))
			^ ((h->flags&NEGATE_DADDR)!=0) ) &&

			(((sp>=h->sport_min)&&(sp<=h->sport_max)))
			^ ((h->flags&NEGATE_SPORT)!=0) &&
			
			(((dp>=h->dport_min)&&(dp<=h->dport_max)))
			^ ((h->flags&NEGATE_DPORT)!=0)){

			if (tcpip_traverse(&tcp_gen, p, l, h->rules))
				return 1;
		}
	}

	return 0;
}

void tcp_match(struct packet *p, unsigned int l)
{
	int i,j;
	int state=0;
	int direction=2;
	u_int32_t flags=p->layer[l+1].flags;

	if ( flags&FLAG_TCP_TRUNC ) return;

	if ( !tcp_stateful ) {
		/* all of the rules are here because we aren't
		 * doing tcp state tracking */
		tcp_do_match(tcp_rules_arr[0][0], p, l);
		return;
	}
	
	if ( !(flags&FLAG_TCP_TRACK) ) {
		/* This happens when we get a truncated TCP header */
		/* printf("tcp: tcp_stateful=on but FLAG_TCP_TRACK is off\n"); */
		goto do_all;
	}

	if ( (flags&FLAG_TCP_STATE) ) {
		state=1;
		if ( flags & FLAG_TCP_SURE ) {
			state=2;
		}
	}

	if ( (flags&FLAG_TCP_2SVR) )
		direction=1;

	for(i=0; i<=state; i++) {
		if ( tcp_do_match(tcp_rules_arr[i][direction], p, l)>0 )
			return;
		
		if ( tcp_do_match(tcp_rules_arr[i][0], p, l)>0 )
			return;
	}

	return;

do_all:
	for(i=0; i<3; i++) {
		for(j=0; j<3; j++) {
			if ( tcp_do_match(tcp_rules_arr[i][j], p, l)>0 )
				return;
		}
	}
}

int tcp_flow(char *sstr, int *sf, int *df)
{
	char str[128];
	char *cur, *tok;
	int end=1;
	
	if ( snprintf(str, sizeof(str), "%s", sstr)<0 )
		return 0;

	for(tok=cur=str; end; cur++) {
		if ( *cur==',' || *cur==0 ) {
			if ( *cur==0 ) end=0;
			*cur=0;
			if ( !strcmp(tok, "to_server") ||
				!strcmp(tok, "from_client")) {
				*df=1;
			}else if ( !strcmp(tok, "to_client") ||
				!strcmp(tok, "from_server") ) {
				*df=2;
			}else if ( !strcmp(tok, "stateless") ) {
				*sf=1;
			}else if ( !strcmp(tok, "established") ) {
				*sf=2;
			}else{
				printf("flow: warning: '%s' "
					"not supported\n", tok);
			}
			tok=cur+1;
		}
	}
	
	return 1;
}

int tcp_commit(struct rule *r)
{
	struct tcp_head **tcph;
	struct criteria *c;
	int s_flags=1; /* stateless, stateful, established */
	int d_flags=0; /* both, to_server, to_client */

	/* Stateful by default. */
	for(c=r->criteria; c; c=c->next) {
		if ( !strcmp("stateless", c->crit) ) {
			s_flags=0;
		}else if ( !strcmp("flow", c->crit) ) {
			if ( !tcp_flow(c->args, &s_flags, &d_flags) )
				return 0;
		}
	}

	if ( tcp_stateful ) {
		tcph=&tcp_rules_arr[s_flags][d_flags];
	}else{
		/* If we aren't doing state tracking then
		 * all rules fall in to the first bucket */
		tcph=&tcp_rules_arr[0][0];
	}

	return tcp_udp_commit(tcph,
		&num_tcp_head,
		&num_tcp_rules, r,
		tcp_ignore_words);
}

struct proto tcp_p=init_proto2("tcp", tcp_decode, tcp_dprint, NULL, tcp_commit);
struct proto_req tcp_r[]={
	proto_request("ip", 6),
	null_request()
};

int tcp_dprint(struct layer *l, char *buf, int buflen)
{
	char tcpflags[]="FSRPAU12";
	char disp[9];
	int n, x;
	
	for(n=0,x=1; n<8; n++, x*=2) {
		if ( l->h.tcp->flags.flags & x ){
			disp[n]=tcpflags[n];
		}else{
			disp[n]='*';
		}
	}disp[8]='\0';
	
	return snprintf(buf, buflen,
		"%u > %u [%s] seq=0x%x ack=0x%x win=%u",
		ntohs(l->h.tcp->sport),
		ntohs(l->h.tcp->dport),
		disp,
		ntohl(l->h.tcp->seq),
		ntohl(l->h.tcp->ack),
		ntohs(l->h.tcp->win));
}

static inline u_int32_t tcp_csum(struct packet *pkt, unsigned int l)
{
	struct pkt_iphdr *iph=pkt->layer[l].h.ip;
	struct pkt_tcphdr *tcph=pkt->layer[l+1].h.tcp;
	struct tcp_phdr ph;
	u_int16_t *tmp;
	u_int32_t sum=0;
	u_int16_t len, csum;
	int i;
	
	if ( pkt->layer[l].flags&FLAG_IP_DTRUNC )
		return 0;

	len=ntohs(iph->tot_len) - (iph->ihl<<2);

	/* Make pseudo-header */
	ph.sip=iph->saddr;
	ph.dip=iph->daddr;
	ph.zero=0;
	ph.proto=iph->protocol;
	ph.tcp_len=htons(len);

	/* Checksum the pseudo-header */
	tmp=(u_int16_t *)&ph;
	for(i=0; i<6; i++) {
		sum+=tmp[i];
		if(sum & 0x80000000) {
			sum = (sum & 0xffff) + (sum >> 16);
		}
	}
	
	/* Checksum the header+data */
	tmp=(u_int16_t *)tcph;
	for(i=0; i<len>>1; i++) {
		sum+=tmp[i];
		if(sum & 0x80000000) {
			sum = (sum & 0xffff) + (sum >> 16);
		}
	}

	/* Deal with last byte (if odd number of bytes) */
	if ( len&1 ) {
		union {
			u_int8_t b[2];
			u_int16_t s;
		}f;

		f.b[0]=((u_int8_t *)tcph)[len-1];
		f.b[1]=0;
		sum+=f.s;
	}

	while(sum >> 16)
		sum = (sum & 0xffff) + (sum >> 16);

	csum=(~sum) & 0xffff;

	return csum ? 0 : FLAG_TCP_CSUM;
}

void tcp_decode(struct packet *p)
{
	struct tcp_session *s;
	struct layer *l=&p->layer[p->llen-1];
	int ks=0;

	/* State tracking */
	if ( tcp_stateful && 
		l->h.raw+sizeof(struct pkt_tcphdr) <= p->end ){
		l->flags|=tcp_csum(p, p->llen-2);
		ks=tcpstream_process(p, p->llen-2);
	}

	if ( p->llen >= PKT_LAYERS ) goto err;

	/* Check it is at least minimum size */
	if ( l->h.raw+sizeof(struct pkt_tcphdr) > p->end ) {
		l->flags|=FLAG_TCP_TRUNC;
		goto err;
	}

	/* Use the doff field to decode the packet */
	if ( (p->layer[p->llen].h.raw=
		l->h.raw+(l->h.tcp->doff<<2))
		> p->end ) {
		p->layer[p->llen].h.raw=p->end;
		goto err;
	}

	/* Can't do it without tcpstream */
	if ( !tcp_stateful ) goto nostate;
	if ( !(s=l->session) ) goto nostate;
	if ( !s->proto ) goto nostate;

	p->layer[p->llen].flags=0;
	p->layer[p->llen].session=NULL;
	p->layer[p->llen++].proto=s->proto;
	s->proto->decode(p);
	if ( ks ) tcp_free(l->session);
	return;

nostate:
	p->layer[p->llen++].proto=NULL;
err:
	preproc_dispatch(p);
	if ( ks ) tcp_free(l->session);
}
