#ifndef __ELOG_HEADER_INCLUDED__
#define __ELOG_HEADER_INCLUDED__

/* 
 * An extended logfile is described by a single header at the start
 * of a file - it contains little more than a magic number.
 * 
 * Each packet is represented as a variable length record comprising of:
 *  o a fixed size header
 *  o a variable number of variable length records describing
 *    each layer in the packet such as protocol name, offset 
 *    within the packet and additional data.
 *  o optional packet data
 *  o an optional variable length generator string
 *  o an optional variable length alert string
 *  o Any extra padding
 *
 * Exact structures are described below.
 *
 */

/* Extended log file header (big endian integers) */
#define EF_MAGIC 0xefeff00f
#define EF_VERS_MIN 1
#define EF_VERS_MAJ 0
struct efile_hdr {
	u_int32_t magic;
	u_int16_t flags;
	u_int8_t vers_minor;
	u_int8_t vers_major;
};

/* Extended log packet header */
struct elog_pkthdr {
	u_int32_t reclen;	/* total size of record */
	u_int32_t sid,rev;	/* signature id and revision */
	u_int32_t pflags;	/* packet flags */
	u_int16_t decode_len;	/* decode data length (multiples of 4) */
	u_int8_t gen_len;	/* generator length (+null) */
	u_int8_t alert_len;	/* alert length (+null) */
	u_int32_t pkt_len;	/* packet length on wire */
	u_int32_t pkt_caplen;	/* length of present captured packet */
	struct timeval ts;	/* time of capture */
	/* <protocol decode data> */
	/* pkt_caplen bytes of data */
	/* null terminated generator string (gen_len) */
	/* null terminated alert string (alert_len) */
	/* any extra padding */
};

/* decode data for each layer */
struct elog_dhdr {
	u_int8_t	tot_len;	/* multiples of 4 */
	u_int8_t	name_len;	/* multiples of 4 */
	u_int16_t 	pkt_ofs;
	u_int32_t	flags;
	/* null terminated protocol name */
	/* protocol private data */
};

#endif /* __ELOG_HEADER_INCLUDED__ */
