#ifndef __PKT_TCP_HEADER_INCLUDED__
#define __PKT_TCP_HEADER_INCLUDED__

#define FLAG_TCP_REASM	0x01 /* Reassembled segment */
#define FLAG_TCP_CSUM	0x02 /* Checksum OK */
#define FLAG_TCP_STATE	0x04 /* Valid stream */
#define FLAG_TCP_SURE	0x08 /* Assured (seen packets in both directions) */
#define FLAG_TCP_TRACK	0x10 /* Conntrack is even on? */
#define FLAG_TCP_2SVR	0x20 /* To server */
#define FLAG_TCP_CT_EST	0x40 /* First packet to be marked FLAG_TCP_SURE */
#define FLAG_TCP_TRUNC	0x80 /* Header is truncated */

/* MISC */
#define TCP_MSS			512
#define TCP_MAXWIN		65535
#define TCP_MAX_WINSHIFT	16

/* OPTIONS */
#define TCPI_OPT_TIMESTAMPS	1
#define TCPI_OPT_SACK		2
#define TCPI_OPT_WSCALE		4
#define TCPI_OPT_ECN		8

/* TCP FLAGS */
#define TCP_FIN		0x01	/* Finish */
#define TCP_SYN		0x02	/* Synchronise */
#define TCP_RST		0x04	/* Reset */
#define TCP_PSH		0x08	/* Push */
#define TCP_ACK		0x10	/* Acknowlege */
#define TCP_URG		0x20	/* Urgent pointer */
#define TCP_ECE		0x40	/* ECN echo */
#define TCP_CWR		0x80	/* Congestion window reduced */

#define TCP_STD		(TCP_FIN|TCP_SYN|TCP_RST|TCP_PSH|TCP_ACK|TCP_URG)

typedef union _tcpflags {
	u_int8_t flags;
	struct {
#if __BYTE_ORDER == __LITTLE_ENDIAN
	u_int8_t fin:1;
	u_int8_t syn:1;
	u_int8_t rst:1;
	u_int8_t psh:1;
	u_int8_t ack:1;
	u_int8_t urg:1;
	u_int8_t ece:1;
	u_int8_t cwr:1;
#elif __BYTE_ORDER == __BIG_ENDIAN
	u_int8_t cwr:1;
	u_int8_t ece:1;
	u_int8_t urg:1;
	u_int8_t ack:1;
	u_int8_t psh:1;
	u_int8_t rst:1;
	u_int8_t syn:1;
	u_int8_t fin:1;
#else
#error "Couldn't determine endianness"
#endif
	}bits;
}tcpflags;

struct pkt_tcphdr
{
	u_int16_t	sport,dport;
	u_int32_t	seq;
	u_int32_t	ack;
#if __BYTE_ORDER == __LITTLE_ENDIAN
	u_int8_t	res1:4;	/* ??? */
	u_int8_t	doff:4;
#elif __BYTE_ORDER == __BIG_ENDIAN
	u_int8_t	doff:4;
	u_int8_t	res1:4;	/* ??? */
#endif
	tcpflags	flags;
	u_int16_t	win;
	u_int16_t	csum;
	u_int16_t	urp;
};

/* Possible TCP states */
enum{
	TCP_ESTABLISHED = 1,
	TCP_SYN_SENT,
	TCP_SYN_RECV,
	TCP_FIN_WAIT1,
	TCP_FIN_WAIT2,
	TCP_TIME_WAIT,
	TCP_CLOSE,
	TCP_CLOSE_WAIT,
	TCP_LAST_ACK,
	TCP_LISTEN,
	TCP_CLOSING,	 /* now a valid state */
	TCP_MAX_STATES /* Leave at the end! */
};

/* TCP pseudo-header for checksumming */
struct tcp_phdr {
	u_int32_t sip, dip;
	u_int8_t zero, proto;
	u_int16_t tcp_len;
};

/* A simplex tcp stream */
struct tcp_stream {
	u_int8_t	state; /* from above enum */
	u_int8_t	tstamp_ok; /* TIMESTAMP seen on SYN packet */
	u_int8_t	wscale_ok; /* Wscale seen on SYN packet */
	u_int8_t	sack_ok; /* SACK seen on syn packet */

	u_int32_t	snd_una; /* first byte we want ack for */
	u_int32_t	snd_nxt; /* next sequence to send */

	u_int32_t	rcv_nxt; /* what we want to recv next */
	u_int32_t	rcv_wnd; /* receiver window */
	u_int32_t	rcv_wup; /* rcv_nxt on last window update */

	u_int32_t	isn; /* equivalent of rfc793 iss */
};

/* A duplex tcp session */
struct tcp_session {
	u_int32_t c_addr, s_addr; /* client/server address */
	u_int16_t c_port, s_port; /* client/server port */
	struct tcp_session *next,**pprev; /* linked list in hash table */
	struct tcp_stream client; /* client stream state */
	struct tcp_stream server; /* server stream state */
	struct tcp_session *prev_time, *next_time; /* LRU */
	struct tcp_session *related; /* parent relation */
	struct proto *proto; /* Application layer protocol carried on this stream */
	int bucket; /* index in hash table */
};

#endif /* __PKT_TCP_HEADER_INCLUDED__ */
