/*
 * This file is part of firestorm NIDS
 * Copyright (c) 2002 Gianni Tedesco
 *
 * 	The Boyer Moore implementation was pretty much entirely
 * 	taken from snorts mstring.c. The original copyright notice
 * 	of which is reproduced below. That covers the functions
 *	bm_skip(), bm_shift() and str_match(). It hardly matters
 *	as both firestorm and snort are GPL anyway :)
 */

/*
** Copyright (C) 1998-2002 Martin Roesch <roesch@sourcefire.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <netinet/in.h>

#include <packet.h>
#include <signature.h>
#include <decode.h>
#include <template.h>
#include <matcher.h>
#include <plugin.h>

struct proto *http_proto=NULL;
proc_decode_proto decode_proto;

struct content_priv {
	char 		*string;
	unsigned int	len;

	unsigned int	depth;
	unsigned int	offset;

	u_int8_t 	flag_depth;
	u_int8_t	flag_uri;

	/* Boyer moore skip/shift tables */
	int		skip[256];
	int		*shift;
};

/* Content matcher (generic string matching) */
void bm_skip(char *x, int m, int *skip) 
{
	int i;
 
	for (i = 0; i < 256; ++i)
		skip[i] = m+1;
	
	while(m)
		skip[(unsigned char)*x++]=m--;
}
 
void bm_shift(char *ptrn, int plen, int *shift)
{
	int *sptr=shift+plen-1;
	char *pptr=ptrn+plen-1;
	char c;

	c=ptrn[plen-1];
	*sptr=1;

	while(sptr-- != shift) {
		char *p1=ptrn+plen-2, *p2, *p3;

		do {
			while(p1>=ptrn && *p1-- !=c );

			p2=ptrn+plen-2;
			p3=p1;

			while(p3>=ptrn &&
				*p3-- == *p2-- &&
				p2 >= pptr);
		}

		while(p3>=ptrn && 
			*p3-- == *p2-- && 
			p2 >= pptr);

		*sptr=shift+plen-sptr+p2-p3;

		pptr--;
	}
}
 
int str_match_regex(struct packet *p, void *priv, unsigned int l, int n)
{
	struct content_priv *sp=priv;
	int skip_stride,shift_stride, p_idx;
	unsigned int hlen;
	unsigned int dsize;
	int literal=0;
	int regexcomp=0;
	char *hs;
	int b_idx;

	l+=2;
	if ( l>=p->llen ) return n^0;

	if ( http_proto && sp->flag_uri ) {
		struct http_session *h;

		if ( p->layer[l].proto!=http_proto )
			return n^0;

		if ( !(h=p->layer[l].session) ) return n^0;
		if ( !(dsize=h->uri_len) ) return n^0;
		if ( !(hs=h->uri) ) return n^0;
		if ( sp->len+sp->offset > dsize ) return n^0;

		hs+=sp->offset;
		if ( sp->flag_depth && (sp->depth < (dsize-sp->offset)) ) {
			hlen=sp->depth;
		}else{
			hlen=(dsize-sp->offset);
		}
	}else{
		/* Find size of searchable area */
		if ( !(dsize=p->end-p->layer[l].h.raw) ) return n^0;

		/* Make sure packet is big enough */
		if ( sp->len+sp->offset > dsize ) return n^0;

		/* Find start of haystack */
		hs=p->layer[l].h.raw+sp->offset;

		/* Find end of haystack */
		if ( sp->flag_depth && (sp->depth < (dsize-sp->offset)) ) {
			hlen=sp->depth;
		}else{
			hlen=(dsize-sp->offset);
		}
	}

	/* Do the search */
	b_idx=sp->len;
	while(b_idx <= hlen ) {
		p_idx=sp->len;

		while(hs[--b_idx]==sp->string[--p_idx]
			|| (sp->string[--p_idx]=='?' && !literal)
			|| (sp->string[--p_idx]=='*' && !literal)
			|| (sp->string[--p_idx]=='\\' && !literal) ) {

			if ( literal )
				literal=0;

			if ( !literal && sp->string[p_idx]=='\\' )
				literal=1;

			if ( sp->string[p_idx]=='*' ) {
				while(p_idx!=0 && sp->string[--p_idx]=='*');

				while(hs[--b_idx]!=sp->string[p_idx]) {
					regexcomp++;
					if ( b_idx==0 ) return n^0;
				}
			}

			if (p_idx==0) return n^1;
			if (b_idx==0) break;
		}

		skip_stride=sp->skip[(unsigned char)hs[b_idx]];
		shift_stride=sp->shift[p_idx];

		/* micro-optimised of max() function */
		b_idx += ( (skip_stride-shift_stride)>0 )
			? skip_stride : shift_stride;
		b_idx += regexcomp;
		regexcomp=0;
	}
	
	return n^0;
}

int str_match(struct packet *p, void *priv, unsigned int l, int n)
{
	struct content_priv *sp=priv;
	int skip_stride,shift_stride, p_idx;
	unsigned int hlen;
	unsigned int dsize;
	char *hs;
	int b_idx;

	l+=2;
	if ( l>=p->llen ) return n^0;

	if ( http_proto && sp->flag_uri ) {
		struct http_session *h;

		if ( p->layer[l].proto!=http_proto )
			return n^0;

		if ( !(h=p->layer[l].session) ) return n^0;
		if ( !(dsize=h->uri_len) ) return n^0;
		if ( !(hs=h->uri) ) return n^0;
		if ( sp->len+sp->offset > dsize ) return n^0;

		hs+=sp->offset;
		if ( sp->flag_depth && (sp->depth < (dsize-sp->offset)) ) {
			hlen=sp->depth;
		}else{
			hlen=(dsize-sp->offset);
		}
	}else{
		/* Find size of searchable area */
		if ( !(dsize=p->end-p->layer[l].h.raw) ) return n^0;

		/* Make sure packet is big enough */
		if ( sp->len+sp->offset > dsize ) return n^0;

		/* Find start of haystack */
		hs=p->layer[l].h.raw+sp->offset;

		/* Find end of haystack */
		if ( sp->flag_depth && (sp->depth < (dsize-sp->offset)) ) {
			hlen=sp->depth;
		}else{
			hlen=(dsize-sp->offset);
		}
	}

	/* Do the search */
	b_idx=sp->len;
	while(b_idx <= hlen ) {
		p_idx=sp->len;

		while((unsigned char)hs[--b_idx]==
			(unsigned char)sp->string[--p_idx]) {
			if (b_idx<0) return n^0;
			if (p_idx==0) return n^1;
		}

		skip_stride=sp->skip[(unsigned char)hs[b_idx]];
		shift_stride=sp->shift[p_idx];

		/* micro-optimised of max() function */
		b_idx += ( (skip_stride-shift_stride)>0 )
			? skip_stride : shift_stride;
	}
	
	return n^0;
}

int str_match_nocase(struct packet *p, void *priv, unsigned int l, int n)
{
	struct content_priv *sp=priv;
	int skip_stride,shift_stride, p_idx;
	unsigned int hlen;
	unsigned int dsize;
	char *hs;
	int b_idx;

	l+=2;
	if ( l>=p->llen ) return n^0;

	if ( http_proto && sp->flag_uri ) {
		struct http_session *h;

		if ( p->layer[l].proto!=http_proto )
			return n^0;

		if ( !(h=p->layer[l].session) ) return n^0;
		if ( !(dsize=h->uri_len) ) return n^0;
		if ( !(hs=h->uri) ) return n^0;
		if ( sp->len+sp->offset > dsize ) return n^0;

		hs+=sp->offset;
		if ( sp->flag_depth && (sp->depth < (dsize-sp->offset)) ) {
			hlen=sp->depth;
		}else{
			hlen=(dsize-sp->offset);
		}
	}else{		
		/* Find size of searchable area */
		if ( !(dsize=p->end-p->layer[l].h.raw) ) return n^0;

		/* Make sure packet is big enough */
		if ( sp->len+sp->offset > dsize ) return n^0;

		/* Find start of haystack */
		hs=p->layer[l].h.raw+sp->offset;

		/* Find end of haystack */
		if ( sp->flag_depth && (sp->depth < (dsize-sp->offset)) ) {
			hlen=sp->depth;
		}else{
			hlen=(dsize-sp->offset);
		}
	}

	/* Do the search */
	b_idx=sp->len;
	while(b_idx <= hlen ) {
		p_idx=sp->len;
		
		while(toupper((unsigned char)hs[--b_idx])==
			(unsigned char)sp->string[--p_idx]) {
			if (b_idx<0) return n^0;
			if (p_idx==0) return n^1;
		}

		skip_stride=sp->skip[toupper((unsigned char)hs[b_idx])];
		shift_stride=sp->shift[p_idx];

		/* micro-optimised of max() function */
		b_idx += ( (skip_stride-shift_stride)>0 )
			? skip_stride : shift_stride;
	}
	
	return n^0;
}

proc_match_match str_validate(char *args, void **priv, struct criteria *m, int mode)
{
	char *t;
	int state=0;
	char bin[3]={0,0,0};
	unsigned int i=0, j=0;
	struct content_priv *p;
	struct criteria *mod;
	proc_match_match mfunc=str_match;
	int escaped=0;

	if ( !http_proto && !(http_proto=decode_proto("http")) ) return NULL;

	if ( !args ) return NULL;

	if ( !(j=strlen(args)) ) return NULL;

	if ( !(p=calloc(1, sizeof(*p))) ) {
		return NULL;
	}

	for(mod=m; mod; mod=mod->next) {
		if ( !strcmp("depth", mod->crit) ) {
			if ( !strtouint(mod->args, &p->depth) ) {
				free(p);
				return NULL;
			}
			p->flag_depth=1;
		}else if ( !strcmp("offset", mod->crit) ) {
			if ( !strtouint(mod->args, &p->offset) ) {
				free(p);
				return NULL;
			}
		}else if ( !strcmp("regex", mod->crit) ) {
			if ( mfunc==str_match_nocase )
				printf("match_str: nocase and "
					"regex mutually exclusive.\n");
			mfunc=str_match_regex;
		}else if ( !strcmp("nocase", mod->crit) ) {
			if ( mfunc==str_match_regex )
				printf("match_str: nocase and "
					"regex mutually exclusive.\n");
			mfunc=str_match_nocase;
		}
	}

	p->flag_uri=mode;

	/* Finished strings can be no bigger than this */
	if ( !(p->string=malloc(j)) ) {
		free(p);
		return NULL;
	}
	
	for(p->len=j,t=args,j=0; *t; t++) {
		if ( state==0 ) { /* Normal string */
			if ( !escaped && *t=='\\' ) {
				escaped=1;
			}else if ( !escaped && *t=='|' ) {
				state=1;
			}else{
				escaped=0;
				if ( mfunc==str_match_nocase ) {
					p->string[j++]=toupper(
						(unsigned char)*t);
				}else{
					p->string[j++]=*t;
				}
			}
		}else if ( state==1 ) { /* Binary data */
			if ( *t=='|' ) {
				state=0;
				continue;
			}
			
			if ( *t==' ' || *t=='\t' ) {
				continue;
			}

			bin[i++]=*t;

			if ( i==2 ) {
				unsigned long int ul;
				char thisbyte;
				i=0;
				ul=strtoul(bin, NULL, 16);

				/* Impossible to be too big cos
				 * 16^2 = 0xff */
				thisbyte=(char )ul&0xff;

				if ( mfunc==str_match_nocase ) {
					p->string[j++]=toupper(
						(unsigned char)thisbyte);
				}else{
					p->string[j++]=thisbyte;
				}
			}
		}
	}

	if ( i ) {
		printf("match_str: string is n and a half bytes!\n");
		free(p->string);
		free(p);
		return NULL;
	}

	/* Shrink the string if there was any
	 * binary data inside of it */
	if ( j!=p->len ) {
		realloc(p->string, j); /* can only be smaller */
		p->len=j;
	}

	/* Perform boyer-moore preprocessing */
	if ( !(p->shift=(int *)malloc(p->len*sizeof(int))) ) {
		free(p->string);
		free(p);
		return NULL;
	}

	bm_shift(p->string, p->len, p->shift);
	bm_skip(p->string, p->len, p->skip);

	*priv=p;
	return mfunc;
}

proc_match_match content_validate(char *args, void **priv, struct criteria *m)
{
	return str_validate(args, priv, m, 0);
}
proc_match_match uricontent_validate(char *args, void **priv, struct criteria *m)
{
	return str_validate(args, priv, m, 1);
}

void str_cleanup(void *priv) {
	struct content_priv *p=priv;

	if ( p ) {
		if ( p->string ) free(p->string);
		free(p);
	}
}

int _PLUGIN_INIT (struct plugin_in *in, struct plugin_out *out)
{
	int ok=0;
	proc_matcher_add matcher_add;
	
	plugin_check(in, out);
	
	PLUGIN_ID("match.str", "String matching routines");
	PLUGIN_VERSION(2, 0);
	PLUGIN_AUTHOR("Gianni Tedesco", "gianni@scaramanga.co.uk");
	PLUGIN_LICENSE("GPL");

	if ( !(matcher_add=in->import("matcher.add")) ) {
		return PLUGIN_ERR_OBJECT;
	}
	if ( !(decode_proto=in->import("decode.proto")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	ok+=matcher_add("content",
		content_validate,
		NULL,
		str_cleanup);
	ok+=matcher_add("uricontent",
		uricontent_validate,
		NULL,
		str_cleanup);
	
	return (ok) ? PLUGIN_ERR_OK : PLUGIN_ERR_FAIL;
}

int _PLUGIN_UNLOAD (int code) {
	return PLUGIN_ERR_OK;
}
