/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
* 	This module handles the alerting subsystem. Although
* 	at the moment it passes all alerts straight on to the
* 	underlying target plugin - it will eventually support
* 	priority queuing so that non-critical alerts can be
* 	deferred.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <packet.h>
#include <cleanup.h>
#include <sensor.h>
#include <capture.h>
#include <capdev.h>
#include <packet.h>
#include <alert.h>
#include <target.h>

/* Default output mechanisms */
struct output	*d_logs=NULL;
struct output	*d_logs_tail=NULL;
unsigned int	d_mode=GEN_DEFAULT;

/* List of all registered alert generators */
struct generator *generators=NULL;
unsigned int num_generators=0;

/* Register an alert generator */
int alert_generator(struct generator *g)
{
	if ( !g || g->next || !g->name ) {
		printf("alert: generator failed to add. "
			"that could really fuck things up\n");
		return 0;
	}

	g->next=generators;
	generators=g;
	num_generators++;
	return 1;
}

/* Find a registered alert generator */
struct generator *alert_find_generator(char *n)
{
	struct generator *g;

	for(g=generators; g; g=g->next) {
		if ( !strcmp(g->name, n) )
			return g;
	}

	return NULL;
}

/* Close and open all logs */
void alert_hup(void)
{
	struct generator *g;
	struct output *o;
	
	for(g=generators; g; g=g->next) {
		for(o=g->logs; o; o=o->next) {
			o->target->hup(o->t_priv);
		}
	}

	for(o=d_logs; o; o=o->next) {
		o->target->hup(o->t_priv);
	}
}

/* Cleanup all open logfiles */
void alert_cleanup(int state, void *priv)
{
	struct generator *g;
	struct output *o;

	printf("alert: flushing all logfiles\n");
	
	for(g=generators; g; g=g->next) {
		for(o=g->logs; o; o=o->next) {
			if ( !o->err ) {
				o->target->cleanup(o->t_priv);
			}
		}
	}

	for(o=d_logs; o; o=o->next) {
		if ( !o->err ) {
			o->target->cleanup(o->t_priv);
		}
	}
}

/* Initialise out cleanup handler */
void alert_init(void)
{
	cleanup_add(alert_cleanup, NULL);
}

/* Config file hook for setting up output mechanisms */
int alert_output(char *gen, char *type, char *args)
{
	struct output *o;
	struct generator *g;

	struct output **logs;
	struct output **logs_tail;
	unsigned int *mode;

	if ( !strcmp(gen, "default") ) {
		logs=&d_logs;
		logs_tail=&d_logs_tail;
		mode=&d_mode;
	}else{
		if( !(g=alert_find_generator(gen)) ) {
			return 0;
		}

		logs=&g->logs;
		logs_tail=&g->logs;
		mode=&g->mode;
	}

	if ( !strcmp(type, "suppress") ) {
		*mode=GEN_SUPPRESS;
	}else{
		if ( !(o=malloc(sizeof(*o))) ) {
			return 0;
		}

		if ( !(o->target=target_find(type)) ) {
			free(o);
			return 0;
		}

		if ( !(o->target->validate(args, &o->t_priv)) ) {
			free(o);
			return 0;
		}

		o->next=NULL;
		o->err=0;
		if ( *logs==NULL ) {
			*logs=o;
			*logs_tail=o;
		}else{
			(*logs_tail)->next=o;
			*logs_tail=o;
		}
	}

	return 1;
}

/* This is where alerts come to */
void alert(struct generator *gen, struct packet *p, struct alert *a)
{
	struct output *l;

	/* Supress this alert */
	if ( gen->mode==GEN_SUPPRESS ) {
		return;
	}
	
	if ( gen->logs==NULL && gen->mode==GEN_DEFAULT ) {
		if ( d_mode==GEN_SUPPRESS ) return;
		l=d_logs;
	}else{
		l=gen->logs;
	}

	/* If it's empty, use builtin target */
	if ( !l ) {
		b_target(gen, p, a, NULL);
		return;
	}

	/* Loop through all specified logs */
	for(;l; l=l->next) {
		if ( !l->err ) {
			l->target->target(gen, p,
				a, l->t_priv);
		}else{
			printf("alert: error sending alert\n");
		}
	}
}
