/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
*    This file represents the capture subsystem. It's really
*    just a bunch of thin wrappers around capdev plugins.
*/
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <cleanup.h>
#include <packet.h>
#include <capture.h>
#include <capdev.h>
#include <alert.h>

struct cap_s capture;
unsigned int num_captures=0;

void capture_cleanup(int state, void *priv)
{
	if ( capture.priv )
		capture.capdev->end(capture.priv);
	if ( capture.args )
		free(capture.args);
}

void capture_stop(void){capture.state=CAP_STATE_STOP;}
void capture_interrupt(void){capture.state=CAP_STATE_INT;}
void capture_init(void) {cleanup_add(capture_cleanup, NULL);}

int capture_add(char *type, char *args)
{
	struct capdev *cd;
	char *a;

	if ( !type || !args ) return 0;

	if ( !(cd=capdev_find(type)) ) {
		return 0;
	}
	
	if ( !(a=strdup(args)) ) {
		cperror("strdup");
	}
	
	capture.capdev=cd;
	capture.args=a;
	if ( !(capture.priv=capture.capdev->init(args)) )
		cleanup(EXIT_ERR, "Capture failed to initialise");
	capture.state=CAP_STATE_INIT;

	num_captures++;

	return 1;
}

/* Go through the first capture */
void capture_go(void)
{
	struct timeval tv, tv2;
	struct timeval end;

	if ( !num_captures ) {
		cleanup(EXIT_ERR, "No captures specified");
	}
	
	/* Ready to go */
	printf("capture: %s[%s]: started\n",
		capture.capdev->name, capture.args);

	fflush(stdout);

	/* Do the capture */
	capture.state=CAP_STATE_CAPTURE;
	gettimeofday(&tv, NULL);
	do {
		capture.capdev->go(capture.priv, &capture);
		if ( capture.state==CAP_STATE_INT ) {
			capture.state=CAP_STATE_CAPTURE;
			alert_hup();
		}
	} while ( capture.state==CAP_STATE_CAPTURE );
	gettimeofday(&tv2, NULL);
	capture.state=CAP_STATE_ENDED;

	/* All done */
	printf("capture: %s[%s]: stopped: %llu packets\n",
		capture.capdev->name,capture.args,the_serial);

	/* Calculate elapsed times (badly) */
	end.tv_sec=(tv2.tv_sec-tv.tv_sec)-1;
	end.tv_usec=(1000000+tv2.tv_usec)-tv.tv_usec;
	while ( end.tv_usec >= 1000000 ) {
		end.tv_usec-=1000000;
		end.tv_sec++;
	}

	printf("  BEGIN: %.lu.%.6lu\n", tv.tv_sec, tv.tv_usec);
	printf("    END: %.lu.%.6lu\n", tv2.tv_sec, tv2.tv_usec);
	printf("ELAPSED: %lu.%.6lu\n", end.tv_sec, end.tv_usec);
}
